from __future__ import absolute_import
import abc
import concurrent.futures
import sys

import shortuuid
import six

from . import futures

__all__ = [
    'Communicator', 'CommunicatorHelper', 'RemoteException', 'DeliveryFailed', 'TaskRejected', 'UnroutableError',
    'TimeoutError', 'DuplicateSubscriberIdentifier'
]


class RemoteException(Exception):
    """An exception occurred at the remote end of the call """


class DeliveryFailed(Exception):
    """Failed to deliver a message """


class UnroutableError(DeliveryFailed):
    """The messages was unroutable """


class TaskRejected(Exception):
    """ A task was rejected at the remote end """


class DuplicateSubscriberIdentifier(Exception):
    """Failed to add a subscriber because the identifier supplied is already in use"""


TimeoutError = concurrent.futures.TimeoutError  # pylint: disable=redefined-builtin


@six.add_metaclass(abc.ABCMeta)
class Communicator(object):
    """
    The interface for a communicator used to both send and receive various
    types of message.
    """

    @abc.abstractmethod
    def add_rpc_subscriber(self, subscriber, identifier=None):
        pass

    @abc.abstractmethod
    def remove_rpc_subscriber(self, identifier):
        """
        Remove an RPC subscriber given the identifier.  Raises a `ValueError` if there
        is no such subscriber.

        :param identifier: The RPC subscriber identifier
        """
        pass

    @abc.abstractmethod
    def add_task_subscriber(self, subscriber):
        pass

    @abc.abstractmethod
    def remove_task_subscriber(self, subscriber):
        pass

    @abc.abstractmethod
    def add_broadcast_subscriber(self, subscriber, identifier=None):
        """
        Add a broadcast subscriber that will receive all broadcast messages

        :param subscriber: the subscriber function to be called
        :param identifier: an optional identifier for the subscriber
        :return: an identifier for the subscriber and can be subsequently used to remove it
        """
        pass

    @abc.abstractmethod
    def remove_broadcast_subscriber(self, identifier):
        """
        Remove a broadcast subscriber

        :param identifier: the identifier of the subscriber to remove
        """
        pass

    @abc.abstractmethod
    def task_send(self, task, no_reply=False):
        """
        Send a task messages, this will be queued and picked up by a
        worker at some point in the future.  The method returns a future
        representing the outcome of the task.

        :param task: The task message
        :param no_reply: Do not send a reply containing the result of the task
        :type no_reply: bool
        :return: A future corresponding to the outcome of the task
        :rtype: :class:`kiwi.Future`
        """

    @abc.abstractmethod
    def rpc_send(self, recipient_id, msg):
        """
        Initiate a remote procedure call on a recipient.  This method
        returns a future representing the outcome of the call.

        :param recipient_id: The recipient identifier
        :param msg: The body of the message
        :return: A future corresponding to the outcome of the call
        :rtype: :class:`kiwi.Future`
        """
        pass

    @abc.abstractmethod
    def broadcast_send(self, body, sender=None, subject=None, correlation_id=None):
        pass


class CommunicatorHelper(Communicator):
    # Have to disable this linter because this class remains abstract and it is
    # just used by calsses that will themselves be concrete
    # pylint: disable=abstract-method

    def __init__(self):
        self._task_subscribers = []
        self._broadcast_subscribers = {}
        self._rpc_subscribers = {}

    def add_rpc_subscriber(self, subscriber, identifier=None):
        identifier = identifier or shortuuid.uuid()
        if identifier in self._rpc_subscribers:
            raise DuplicateSubscriberIdentifier("RPC identifier '{}'".format(identifier))
        self._rpc_subscribers[identifier] = subscriber

    def remove_rpc_subscriber(self, identifier):
        try:
            self._rpc_subscribers.pop(identifier)
        except KeyError:
            raise ValueError("Unknown subscriber '{}'".format(identifier))

    def add_task_subscriber(self, subscriber):
        """
        Register a task subscriber

        :param subscriber: The task callback function
        """
        self._task_subscribers.append(subscriber)

    def remove_task_subscriber(self, subscriber):
        """
        Remove a task subscriber

        :param subscriber: The task callback function
        """
        try:
            self._task_subscribers.remove(subscriber)
        except ValueError:
            raise ValueError("Unknown subscriber: '{}'".format(subscriber))

    def add_broadcast_subscriber(self, subscriber, identifier=None):
        identifier = identifier or shortuuid.uuid()
        if identifier in self._broadcast_subscribers:
            raise DuplicateSubscriberIdentifier("Broadcast identifier '{}'".format(identifier))

        self._broadcast_subscribers[identifier] = subscriber
        return identifier

    def remove_broadcast_subscriber(self, identifier):
        try:
            del self._broadcast_subscribers[identifier]
        except KeyError:
            raise ValueError("Broadcast subscriber '{}' unknown".format(identifier))

    def fire_task(self, msg, no_reply=False):
        future = futures.Future()
        handled = False

        for subscriber in self._task_subscribers:
            try:
                result = subscriber(self, msg)
                future.set_result(result)
                handled = True
                break
            except TaskRejected:
                pass
            except Exception:  # pylint: disable=broad-except
                future.set_exception(RemoteException(sys.exc_info()))
                handled = True
                break

        if not handled:
            future.set_exception(TaskRejected("Rejected by all subscribers"))

        if no_reply:
            return None

        return future

    def fire_rpc(self, recipient_id, msg):
        try:
            subscriber = self._rpc_subscribers[recipient_id]
        except KeyError:
            raise UnroutableError("Unknown rpc recipient '{}'".format(recipient_id))
        else:
            future = futures.Future()
            try:
                future.set_result(subscriber(self, msg))
            except Exception:  # pylint: disable=broad-except
                future.set_exception(RemoteException(sys.exc_info()))

            return future

    def fire_broadcast(self, body, sender=None, subject=None, correlation_id=None):
        for subscriber in self._broadcast_subscribers.values():
            subscriber(self, body=body, sender=sender, subject=subject, correlation_id=correlation_id)
        return True



from __future__ import absolute_import
import collections
import logging
import uuid
import sys
import traceback

from tornado import gen, concurrent
import topika

import kiwipy
from . import defaults
from . import messages
from . import utils

_LOGGER = logging.getLogger(__name__)

__all__ = ['RmqTaskSubscriber', 'RmqTaskPublisher']

TaskBody = collections.namedtuple('TaskBody', ('task', 'no_reply'))


class RmqTaskSubscriber(messages.BaseConnectionWithExchange):
    """
    Listens for tasks coming in on the RMQ task queue
    """
    TASK_QUEUE_ARGUMENTS = {"x-message-ttl": defaults.TASK_MESSAGE_TTL}

    def __init__(self,
                 connection,
                 exchange_name=defaults.MESSAGE_EXCHANGE,
                 queue_name=defaults.TASK_QUEUE,
                 testing_mode=False,
                 decoder=defaults.DECODER,
                 encoder=defaults.ENCODER,
                 exchange_params=None,
                 prefetch_size=defaults.TASK_PREFETCH_SIZE,
                 prefetch_count=defaults.TASK_PREFETCH_COUNT):
        # pylint: disable=too-many-arguments
        """
        :param connection: An RMQ connection
        :type connection: :class:`topika.Connection`
        :param exchange_name: the name of the exchange to use
        :type exchange_name: :class:`six.string_types`
        :param queue_name: the name of the task queue to use
        :type queue_name: :class:`six.string_types`
        :param decoder: A message decoder
        :param encoder: A response encoder
        """
        super(RmqTaskSubscriber, self).__init__(
            connection, exchange_name=exchange_name, exchange_params=exchange_params, testing_mode=testing_mode)

        self._task_queue_name = queue_name
        self._testing_mode = testing_mode
        self._decode = decoder
        self._encode = encoder
        self._prefetch_size = prefetch_size
        self._prefetch_count = prefetch_count
        self._consumer_tag = None

        self._task_queue = None  # type: topika.Queue
        self._subscribers = []
        self._pending_tasks = []

    def add_task_subscriber(self, subscriber):
        self._subscribers.append(subscriber)
        if self._consumer_tag is None:
            self._consumer_tag = self._task_queue.consume(self._on_task)

    def remove_task_subscriber(self, subscriber):
        self._subscribers.remove(subscriber)
        if not self._subscribers:
            self._task_queue.cancel(self._consumer_tag)
            self._consumer_tag = None

    @gen.coroutine
    def connect(self):
        if self.channel():
            # Already connected
            return

        yield super(RmqTaskSubscriber, self).connect()
        yield self.channel().set_qos(prefetch_count=self._prefetch_count, prefetch_size=self._prefetch_size)

        yield self._create_task_queue()

    @gen.coroutine
    def _create_task_queue(self):
        """Create and bind the task queue"""
        arguments = dict(self.TASK_QUEUE_ARGUMENTS)
        if self._testing_mode:
            arguments['x-expires'] = defaults.TEST_QUEUE_EXPIRES

        # x-expires means how long does the queue stay alive after no clients
        # x-message-ttl means what is the default ttl for a message arriving in the queue
        self._task_queue = yield self._channel.declare_queue(
            name=self._task_queue_name,
            durable=not self._testing_mode,
            auto_delete=self._testing_mode,
            arguments=arguments)
        yield self._task_queue.bind(self._exchange, routing_key=self._task_queue.name)

    @gen.coroutine
    def _on_task(self, message):
        """
        :param message: The topika RMQ message
        :type message: :class:`topika.IncomingMessage`
        """
        # pylint: disable=too-many-branches

        with message.process(ignore_processed=True):
            handled = False
            # Decode the message tuple into a task body for easier use
            task_body = TaskBody(*self._decode(message.body))  # type: TaskBody
            for subscriber in self._subscribers:
                try:
                    subscriber = utils.ensure_coroutine(subscriber)
                    result = yield subscriber(self, task_body.task)

                    # If a task returns a future it is not considered done until the chain of futures
                    # (i.e. if the first future resolves to a future and so on) finishes and produces
                    # a concrete result
                    while concurrent.is_future(result):
                        if not task_body.no_reply:
                            yield self._send_response(utils.pending_response(), message)
                        result = yield result
                except kiwipy.TaskRejected:
                    # Keep trying to find one that will accept the task
                    continue
                except kiwipy.CancelledError as exception:
                    if not task_body.no_reply:
                        reply_body = utils.cancelled_response(str(exception))
                    handled = True  # Finished
                except KeyboardInterrupt:  # pylint: disable=try-except-raise
                    raise
                except Exception as exc:  # pylint: disable=broad-except
                    if task_body.no_reply:
                        # The user has asked for no reply so log an error so they see this exception
                        _LOGGER.exception('The task excepted')
                    else:
                        # Send the exception back to the other side but log here at INFO level also
                        reply_body = utils.exception_response(sys.exc_info()[1:])
                        _LOGGER.info('There was an exception in a task %s:\n%s', exc, traceback.format_exc())
                    handled = True  # Finished
                else:
                    # Create a reply message
                    if not task_body.no_reply:
                        reply_body = utils.result_response(result)
                    handled = True  # Finished

                if handled:
                    message.ack()
                    if not task_body.no_reply:
                        yield self._send_response(reply_body, message)
                    break  # Done, do break the loop

            if not handled:
                # No one handled the task
                message.reject(requeue=True)

    def _build_response_message(self, body, incoming_message):
        """
        Create a topika Message as a response to a task being deal with.

        :param body: The message body dictionary
        :type body: dict
        :param incoming_message: The original message we are responding to
        :type incoming_message: :class:`topika.IncomingMessage`
        :return: The response message
        :rtype: :class:`topika.Message`
        """
        # Add host info
        body[utils.HOST_KEY] = utils.get_host_info()
        message = topika.Message(body=self._encode(body), correlation_id=incoming_message.correlation_id)

        return message

    @gen.coroutine
    def _send_response(self, msg_body, incoming_message):
        msg = self._build_response_message(msg_body, incoming_message)
        yield self._exchange.publish(msg, routing_key=incoming_message.reply_to)


class RmqTaskPublisher(messages.BasePublisherWithReplyQueue):
    """
    Publishes messages to the RMQ task queue and gets the response
    """

    def __init__(self,
                 connection,
                 task_queue_name=defaults.TASK_QUEUE,
                 exchange_name=defaults.MESSAGE_EXCHANGE,
                 exchange_params=None,
                 encoder=defaults.ENCODER,
                 decoder=defaults.DECODER,
                 confirm_deliveries=True,
                 testing_mode=False):
        # pylint: disable=too-many-arguments
        super(RmqTaskPublisher, self).__init__(
            connection,
            exchange_name=exchange_name,
            exchange_params=exchange_params,
            encoder=encoder,
            decoder=decoder,
            confirm_deliveries=confirm_deliveries,
            testing_mode=testing_mode)
        self._task_queue_name = task_queue_name

    @gen.coroutine
    def task_send(self, task, no_reply=False):
        """
        Send a task for processing by a task subscriber
        :param task: The task payload
        :param no_reply: Don't send a reply containing the result of the task
        :type no_reply: bool
        :return: A future representing the result of the task
        :rtype: :class:`tornado.concurrent.Future`
        """
        # Build the full message body and encode as a tuple
        body = self._encode((task, no_reply))
        # Now build up the full topika message
        task_msg = topika.Message(body=body, correlation_id=str(uuid.uuid4()), reply_to=self._reply_queue.name)

        result_future = None
        if no_reply:
            published = yield self.publish(task_msg, routing_key=self._task_queue_name, mandatory=True)
        else:
            published, result_future = yield self.publish_expect_response(
                task_msg, routing_key=self._task_queue_name, mandatory=True)

        assert published, "The task was not published to the exchange"
        raise gen.Return(result_future)



"""
A generic test for any communicator that conforms to the `kiwipy.Communicator` interface.

I've had to disable invalid-name for some portions of this file as I use the convention
that all caps 'variable' names mean that it's a constant.
"""

from __future__ import absolute_import
import abc
import six

import kiwipy


@six.add_metaclass(abc.ABCMeta)
class CommunicatorTester(object):
    # Disable invalid name because I use caps for constants which the linter doesn't like
    # also disable no-member for superclass calls because we use this as a mixin that gets used
    # with unittest.TestCase
    # pylint: disable=invalid-name, no-member

    WAIT_TIMEOUT = 5.

    communicator = None

    def setUp(self):
        super(CommunicatorTester, self).setUp()
        self.communicator = self.create_communicator()

    def tearDown(self):
        self.destroy_communicator(self.communicator)
        self.communicator = None

    @abc.abstractmethod
    def create_communicator(self):
        """
        :return: A constructed communicator
        :rtype: :class:`kiwi.Communicator`
        """
        pass

    def destroy_communicator(self, communicator):
        pass

    # region RPC

    def test_rpc_send_receive(self):
        MESSAGE = "sup yo'"  # pylint: disable=invalid-name
        RESPONSE = "nuthin bra"  # pylint: disable=invalid-name

        messages = []

        def on_receive(_comm, msg):
            messages.append(msg)
            return RESPONSE

        self.communicator.add_rpc_subscriber(on_receive, 'rpc')
        response = self.communicator.rpc_send('rpc', MESSAGE).result(timeout=self.WAIT_TIMEOUT)

        self.assertEqual(MESSAGE, messages[0])
        self.assertEqual(RESPONSE, response)

    def test_add_remove_rpc_subscriber(self):
        """ Test adding, sending to, removing and readding an RPC subscriber """

        def rpc_subscriber(_comm, _msg):
            return True

        # Check we're getting messages
        self.communicator.add_rpc_subscriber(rpc_subscriber, rpc_subscriber.__name__)
        result = self.communicator.rpc_send(rpc_subscriber.__name__, None).result(timeout=self.WAIT_TIMEOUT)
        self.assertTrue(result)

        self.communicator.remove_rpc_subscriber(rpc_subscriber.__name__)
        # Check that we're unsubscribed
        with self.assertRaises((kiwipy.UnroutableError, kiwipy.TimeoutError)):
            self.communicator.rpc_send(rpc_subscriber.__name__, None).result(timeout=self.WAIT_TIMEOUT)

        self.communicator.add_rpc_subscriber(rpc_subscriber, rpc_subscriber.__name__)
        result = self.communicator.rpc_send(rpc_subscriber.__name__, None).result(timeout=self.WAIT_TIMEOUT)
        self.assertTrue(result)

    def test_rpc_nested_futrues(self):
        """Test that an RPC call that returns a future, which itself resolves to a future works"""
        RESULT = "You've reached the bottom of the rabbit hole my friend"  # pylint: disable=invalid-name

        def rpc_subscriber(_comm, _msg):
            future1 = kiwipy.Future()
            future2 = kiwipy.Future()
            future1.set_result(future2)
            future2.set_result(RESULT)
            return future1

        # Check we're getting messages
        self.communicator.add_rpc_subscriber(rpc_subscriber, rpc_subscriber.__name__)
        future1 = self.communicator.rpc_send(rpc_subscriber.__name__, None).result(timeout=self.WAIT_TIMEOUT)
        self.assertIsInstance(future1, kiwipy.Future)

        future2 = future1.result(timeout=self.WAIT_TIMEOUT)
        self.assertIsInstance(future2, kiwipy.Future)

        result = future2.result(timeout=self.WAIT_TIMEOUT)
        self.assertEqual(RESULT, result)

    def test_rpc_exception(self):
        MESSAGE = 'The meaning?'  # pylint: disable=invalid-name
        EXCEPTION = RuntimeError("I cannea do it Captain!")  # pylint: disable=invalid-name

        messages = []

        def on_receive(_com, msg):
            messages.append(msg)
            raise EXCEPTION

        self.communicator.add_rpc_subscriber(on_receive, on_receive.__name__)
        with self.assertRaises(kiwipy.RemoteException):
            self.communicator.rpc_send(on_receive.__name__, MESSAGE).result(timeout=self.WAIT_TIMEOUT)

        self.assertEqual(messages[0], MESSAGE)

    def test_rpc_unroutable(self):
        """Test that if there is no subscriber then an RPC call is unroutable"""
        with self.assertRaises(kiwipy.UnroutableError):
            self.communicator.rpc_send('noone', 'nothing').result(timeout=1.)

    # endregion

    # region Task

    def test_task_send(self):
        TASK = 'The meaning?'  # pylint: disable=invalid-name
        RESULT = 42  # pylint: disable=invalid-name

        tasks = []

        def on_task(_comm, task):
            tasks.append(task)
            return RESULT

        self.communicator.add_task_subscriber(on_task)
        result = self.communicator.task_send(TASK).result(timeout=self.WAIT_TIMEOUT)

        self.assertEqual(tasks[0], TASK)
        self.assertEqual(RESULT, result)

    def test_future_task(self):
        """
        Test a task that returns a future meaning that will be resolve to a value later
        """
        TASK = 'The meaning?'
        RESULT = 42
        result_future = kiwipy.Future()

        tasks = []

        def on_task(_comm, task):
            tasks.append(task)
            return result_future

        self.communicator.add_task_subscriber(on_task)
        task_future = self.communicator.task_send(TASK).result(timeout=self.WAIT_TIMEOUT)

        result_future.set_result(42)

        result = task_future.result(timeout=self.WAIT_TIMEOUT)

        self.assertEqual(TASK, tasks[0])
        self.assertEqual(RESULT, result)

    def test_task_exception(self):
        TASK = 'The meaning?'  # pylint: disable=invalid-name
        EXCEPTION = RuntimeError("I cannea do it Captain!")  # pylint: disable=invalid-name

        tasks = []

        def on_task(_com, task):
            tasks.append(task)
            raise EXCEPTION

        self.communicator.add_task_subscriber(on_task)
        with self.assertRaises(kiwipy.RemoteException):
            self.communicator.task_send(TASK).result(timeout=self.WAIT_TIMEOUT)

        self.assertEqual(tasks[0], TASK)

    def test_task_no_reply(self):
        """Test that we don't get a reply if we don't ask for one, i.e. fire-and-forget"""
        TASK = 'The meaning?'  # pylint: disable=invalid-name
        RESULT = 42  # pylint: disable=invalid-name

        tasks = []

        task_future = kiwipy.Future()

        def on_task(_comm, task):
            tasks.append(task)
            task_future.set_result(RESULT)
            return RESULT

        self.communicator.add_task_subscriber(on_task)
        send_result = self.communicator.task_send(TASK, no_reply=True)
        self.assertIsNone(send_result)

        # Make sure the task is done
        result = task_future.result(timeout=self.WAIT_TIMEOUT)

        self.assertEqual(1, len(tasks))
        self.assertEqual(TASK, tasks[0])
        self.assertEqual(RESULT, result)

    # endregion

    # region Broadcast

    def test_broadcast_send(self):
        SUBJECT = 'yo momma'  # pylint: disable=invalid-name
        BODY = 'so fat'  # pylint: disable=invalid-name
        SENDER_ID = 'me'  # pylint: disable=invalid-name
        FULL_MSG = {'body': BODY, 'subject': SUBJECT, 'sender': SENDER_ID, 'correlation_id': None}  # pylint: disable=invalid-name

        message1 = kiwipy.Future()
        message2 = kiwipy.Future()

        def on_broadcast_1(_comm, body, sender, subject, correlation_id):
            message1.set_result({'body': body, 'subject': subject, 'sender': sender, 'correlation_id': correlation_id})

        def on_broadcast_2(_comm, body, sender, subject, correlation_id):
            message2.set_result({'body': body, 'subject': subject, 'sender': sender, 'correlation_id': correlation_id})

        self.communicator.add_broadcast_subscriber(on_broadcast_1)
        self.communicator.add_broadcast_subscriber(on_broadcast_2)

        self.communicator.broadcast_send(**FULL_MSG)
        # Wait fot the send and receive
        kiwipy.wait((message1, message2), timeout=self.WAIT_TIMEOUT)

        self.assertDictEqual(message1.result(), FULL_MSG)
        self.assertDictEqual(message2.result(), FULL_MSG)

    def test_broadcast_filter_subject(self):
        subjects = []
        EXPECTED_SUBJECTS = ['purchase.car', 'purchase.piano']  # pylint: disable=invalid-name

        done = kiwipy.Future()

        def on_broadcast_1(_comm, _body, _sender=None, subject=None, _correlation_id=None):
            subjects.append(subject)
            if len(subjects) == len(EXPECTED_SUBJECTS):
                done.set_result(True)

        self.communicator.add_broadcast_subscriber(kiwipy.BroadcastFilter(on_broadcast_1, subject="purchase.*"))

        for subj in ['purchase.car', 'purchase.piano', 'sell.guitar', 'sell.house']:
            self.communicator.broadcast_send(None, subject=subj)

        done.result(timeout=self.WAIT_TIMEOUT)

        self.assertEqual(len(subjects), 2)
        self.assertListEqual(EXPECTED_SUBJECTS, subjects)

    def test_broadcast_filter_sender(self):
        EXPECTED_SENDERS = ['bob.jones', 'alice.jones']  # pylint: disable=invalid-name
        senders = []

        done = kiwipy.Future()

        def on_broadcast_1(_comm, _body, sender=None, _subject=None, _correlation_id=None):
            senders.append(sender)
            if len(senders) == len(EXPECTED_SENDERS):
                done.set_result(True)

        self.communicator.add_broadcast_subscriber(kiwipy.BroadcastFilter(on_broadcast_1, sender="*.jones"))

        for sendr in ['bob.jones', 'bob.smith', 'martin.uhrin', 'alice.jones']:
            self.communicator.broadcast_send(None, sender=sendr)

        done.result(timeout=self.WAIT_TIMEOUT)

        self.assertEqual(2, len(senders))
        self.assertListEqual(EXPECTED_SENDERS, senders)

    def test_broadcast_filter_sender_and_subject(self):
        # pylint: disable=invalid-name
        senders_and_subects = set()
        EXPECTED = {
            ('bob.jones', 'purchase.car'),
            ('bob.jones', 'purchase.piano'),
            ('alice.jones', 'purchase.car'),
            ('alice.jones', 'purchase.piano'),
        }

        done = kiwipy.Future()

        def on_broadcast_1(_communicator, _body, sender=None, subject=None, _correlation_id=None):
            senders_and_subects.add((sender, subject))
            if len(senders_and_subects) == len(EXPECTED):
                done.set_result(True)

        filtered = kiwipy.BroadcastFilter(on_broadcast_1)
        filtered.add_sender_filter("*.jones")
        filtered.add_subject_filter("purchase.*")
        self.communicator.add_broadcast_subscriber(filtered)

        for sendr in ['bob.jones', 'bob.smith', 'martin.uhrin', 'alice.jones']:
            for subj in ['purchase.car', 'purchase.piano', 'sell.guitar', 'sell.house']:
                self.communicator.broadcast_send(None, sender=sendr, subject=subj)

        done.result(timeout=self.WAIT_TIMEOUT)

        self.assertEqual(4, len(senders_and_subects))
        self.assertSetEqual(EXPECTED, senders_and_subects)

    def test_add_remove_broadcast_subscriber(self):
        """Test adding, removing and readding a broadcast subscriber"""
        broadcast_received = kiwipy.Future()

        def broadcast_subscriber(_comm, body, sender=None, subject=None, correlation_id=None):
            # pylint: disable=unused-argument
            broadcast_received.set_result(True)

        # Check we're getting messages
        self.communicator.add_broadcast_subscriber(broadcast_subscriber, broadcast_subscriber.__name__)
        self.communicator.broadcast_send(None)
        self.assertTrue(broadcast_received.result(timeout=self.WAIT_TIMEOUT))

        self.communicator.remove_broadcast_subscriber(broadcast_subscriber.__name__)
        # Check that we're unsubscribed
        broadcast_received = kiwipy.Future()
        with self.assertRaises(kiwipy.TimeoutError):
            broadcast_received.result(timeout=self.WAIT_TIMEOUT)

        broadcast_received = kiwipy.Future()
        self.communicator.add_broadcast_subscriber(broadcast_subscriber, broadcast_subscriber.__name__)
        self.communicator.broadcast_send(None)
        self.assertTrue(broadcast_received.result(timeout=self.WAIT_TIMEOUT))

    def test_broadcast_noone_listening(self):
        """ Check that it's not an error to send a broadcast if no one is listening"""
        result = self.communicator.broadcast_send('into the ether')
        self.assertIs(True, result)

    # endregion



from __future__ import absolute_import
import re
import typing
import six

__all__ = ['BroadcastFilter']


class BroadcastFilter(object):

    def __init__(self, subscriber, subject=None, sender=None):
        self._subscriber = subscriber
        self._subject_filters = []
        self._sender_filters = []
        if subject is not None:
            self.add_subject_filter(subject)
        if sender is not None:
            self.add_sender_filter(sender)

    @property
    def __name__(self):
        return 'BroadcastFilter'

    def __call__(self, communicator, body, sender=None, subject=None, correlation_id=None):
        if subject is not None and self._subject_filters and \
                not any([check(subject) for check in self._subject_filters]):
            return

        if sender is not None and self._sender_filters and \
                not any([check(sender) for check in self._sender_filters]):
            return

        self._subscriber(communicator, body, sender, subject, correlation_id)

    def add_subject_filter(self, subject_filter):
        self._subject_filters.append(self._ensure_filter(subject_filter))

    def add_sender_filter(self, sender_filter):
        self._sender_filters.append(self._ensure_filter(sender_filter))

    @classmethod
    def _ensure_filter(cls, filter_value):
        if isinstance(filter_value, six.string_types):
            return re.compile(filter_value.replace('.', '[.]').replace('*', '.*')).match
        if isinstance(filter_value, typing.re.Pattern):
            return filter_value.match

        return lambda val: val == filter_value

    @classmethod
    def _make_regex(cls, filter_str):
        """
        :param filter_str: The filter string
        :type filter_str: basestring
        :return: The regular expression object
        """
        return re.compile(filter_str.replace('.', '[.]'))



from __future__ import absolute_import
import collections
import inspect
import os
import socket
import tornado.concurrent
from tornado import gen, ioloop

import kiwipy

__all__ = []

# The key used in messages to give information about the host that send a message
HOST_KEY = 'host'
HOSTNAME_KEY = 'hostname'
PID_KEY = 'pid'
RESULT_KEY = 'result'
EXCEPTION_KEY = 'exception'
CANCELLED_KEY = 'cancelled'
PENDING_KEY = 'pending'


def get_host_info():
    return {'hostname': socket.gethostname(), 'pid': os.getpid()}


def add_host_info(msg):
    if HOST_KEY in msg:
        raise ValueError("Host information key already exists in message")

    msg[HOST_KEY] = get_host_info()


def result_response(result):
    return {RESULT_KEY: result}


def exception_response(exception, trace=None):
    """
    Create an exception response dictionary
    :param exception: The exception to encode
    :type exception: :class:`Exception`
    :param trace: Optional traceback
    :return: A response dictionary
    :rtype: dict
    """
    msg = str(exception)
    if trace is not None:
        msg += "\n{}".format(trace)
    return {EXCEPTION_KEY: msg}


def cancelled_response(msg=None):
    return {CANCELLED_KEY: msg}


def pending_response(msg=None):
    return {PENDING_KEY: msg}


def response_result(response):
    future = tornado.concurrent.Future()
    response_to_future(response, future)
    return future.result()


def response_to_future(response, future=None):
    """
    Take a response message and set the appropriate value on the given future
    :param response:
    :param future:
    :return:
    """
    if not isinstance(response, collections.Mapping):
        raise TypeError("Response must be a mapping")

    if future is None:
        future = tornado.concurrent.Future()

    if CANCELLED_KEY in response:
        future.cancel()
    elif EXCEPTION_KEY in response:
        future.set_exception(kiwipy.RemoteException(response[EXCEPTION_KEY]))
    elif RESULT_KEY in response:
        future.set_result(response[RESULT_KEY])
    elif PENDING_KEY in response:
        future.set_result(tornado.concurrent.Future())
    else:
        raise ValueError("Unknown response type '{}'".format(response))

    return future


def future_to_response(future):
    """
    Convert a future to a response dictionary
    :param future: The future
    :type future: :class:`kiwipy.Future`
    :return: The response dictionary
    :rtype: dict
    """
    if future.cancelled():
        return cancelled_response()
    try:
        return result_response(future.result())
    except Exception as exception:  # pylint: disable=broad-except
        return exception_response(exception)


def ensure_coroutine(coro_or_fn):
    if gen.is_coroutine_function(coro_or_fn):
        return coro_or_fn
    if callable(coro_or_fn):
        if inspect.isclass(coro_or_fn):
            coro_or_fn = coro_or_fn.__call__
        return gen.coroutine(coro_or_fn)

    raise TypeError('coro_or_fn must be a callable')


@gen.convert_yielded.register(kiwipy.Future)
def _register_concurrent_future(concurrent_future):
    """Register concurrent.future.Future to be compatible with tornado yield"""
    tornado_future = tornado.concurrent.Future()
    tornado.concurrent.chain_future(concurrent_future, tornado_future)
    return tornado_future


def create_task(coro, loop=None):
    """
    Schedule a call to a coroutine in the event loop and wrap the outcome
    in a future.

    :param coro: the coroutine to schedule
    :param loop: the event loop to schedule it in
    :return: the future representing the outcome of the coroutine
    :rtype: :class:`tornado.concurrent.Future`
    """
    loop = loop or ioloop.IOLoop.current()

    future = tornado.concurrent.Future()

    @gen.coroutine
    def run_task():
        with kiwipy.capture_exceptions(future):
            future.set_result((yield coro()))

    loop.add_callback(run_task)
    return future



from __future__ import absolute_import
from __future__ import print_function
from concurrent.futures import Future as ThreadFuture
from functools import partial
import copy
import logging
import threading

import pika
import pika.exceptions
import shortuuid
from tornado import gen, concurrent, ioloop
import topika

import kiwipy
from . import defaults
from . import tasks
from . import messages
from . import utils

__all__ = ['RmqCommunicator', 'RmqThreadCommunicator', 'connect']

_LOGGER = logging.getLogger(__name__)

# The exchange properties use by the publisher and subscriber.  These have to match
# which is why they're declare her
EXCHANGE_PROPERTIES = {'type': topika.ExchangeType.TOPIC}


class RmqPublisher(messages.BasePublisherWithReplyQueue):
    """
    Publisher for sending a range of message types over RMQ
    """
    DEFAULT_EXCHANGE_PARAMS = EXCHANGE_PROPERTIES

    @gen.coroutine
    def rpc_send(self, recipient_id, msg):
        message = topika.Message(body=self._encode(msg), reply_to=self._reply_queue.name)
        published, response_future = yield self.publish_expect_response(
            message, routing_key="{}.{}".format(defaults.RPC_TOPIC, recipient_id), mandatory=True)
        assert published, "The message was not published to the exchanges"
        raise gen.Return(response_future)

    @gen.coroutine
    def broadcast_send(self, msg, sender=None, subject=None, correlation_id=None):
        message_dict = messages.BroadcastMessage.create(
            body=msg,
            sender=sender,
            subject=subject,
            correlation_id=correlation_id,
        )
        message = topika.Message(
            body=self._encode(message_dict),
            delivery_mode=topika.DeliveryMode.NOT_PERSISTENT,
        )
        try:
            result = yield self.publish(message, routing_key=defaults.BROADCAST_TOPIC)
        except pika.exceptions.UnroutableError as exception:
            _LOGGER.debug('Broadcast message sent but was unroutable, probably no queues: %s', str(exception))
            # Doesn't matter because broadcasts are meant to be fire-and-forget anyway
            result = True
        raise gen.Return(result)


class RmqSubscriber(object):
    """
    Subscriber for receiving a range of messages over RMQ
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self,
                 connection,
                 message_exchange=defaults.MESSAGE_EXCHANGE,
                 queue_expires=defaults.QUEUE_EXPIRES,
                 decoder=defaults.DECODER,
                 encoder=defaults.ENCODER,
                 testing_mode=False):
        # pylint: disable=too-many-arguments
        """
        Subscribes and listens for process control messages and acts on them
        by calling the corresponding methods of the process manager.

        :param connection: The tokpia connection
        :type connection: :class:`topika.Connection`
        :param message_exchange: The name of the exchange to use
        :param queue_expires: the expiry time for standard queues in milliseconds.  This is the time after which, if
            there are no subscribers, a queue will automatically be deleted by RabbitMQ.
        :type queue_expires: int
        :param encoder: The encoder to call for encoding a message
        :param decoder: The decoder to call for decoding a message
        :param testing_mode: Run in testing mode: all queues and exchanges
            will be temporary
        """
        super(RmqSubscriber, self).__init__()

        self._connection = connection
        self._channel = None  # type: topika.Channel
        self._exchange = None  # type: topika.Exchange
        self._exchange_name = message_exchange
        self._decode = decoder
        self._testing_mode = testing_mode
        self._response_encode = encoder

        self._broadcast_queue_arguments = {'x-message-ttl': defaults.MESSAGE_TTL}

        self._rmq_queue_arguments = {'x-message-ttl': defaults.MESSAGE_TTL}
        if queue_expires:
            self._rmq_queue_arguments['x-expires'] = queue_expires

        self._rpc_subscribers = {}
        self._broadcast_subscribers = {}
        self._broadcast_queue = None  # type: topika.Queue
        self._broadcast_consumer_tag = None

        self._active = False

    @gen.coroutine
    def add_rpc_subscriber(self, subscriber, identifier=None):
        # Create an RPC queue
        rpc_queue = yield self._channel.declare_queue(exclusive=True, arguments=self._rmq_queue_arguments)
        try:
            identifier = yield rpc_queue.consume(partial(self._on_rpc, subscriber), consumer_tag=identifier)
        except pika.exceptions.DuplicateConsumerTag:
            raise kiwipy.DuplicateSubscriberIdentifier("RPC identifier '{}'".format(identifier))
        else:
            yield rpc_queue.bind(self._exchange, routing_key='{}.{}'.format(defaults.RPC_TOPIC, identifier))
            # Save the queue so we can cancel and unbind later
            self._rpc_subscribers[identifier] = rpc_queue
            raise gen.Return(identifier)

    @gen.coroutine
    def remove_rpc_subscriber(self, identifier):
        try:
            rpc_queue = self._rpc_subscribers.pop(identifier)
        except KeyError:
            raise ValueError("Unknown subscriber '{}'".format(identifier))
        else:
            yield rpc_queue.cancel(identifier)
            yield rpc_queue.unbind(self._exchange, routing_key='{}.{}'.format(defaults.RPC_TOPIC, identifier))

    @gen.coroutine
    def add_broadcast_subscriber(self, subscriber, identifier=None):
        identifier = identifier or shortuuid.uuid()
        if identifier in self._broadcast_subscribers:
            raise kiwipy.DuplicateSubscriberIdentifier("Broadcast identifier '{}'".format(identifier))

        self._broadcast_subscribers[identifier] = subscriber
        if self._broadcast_consumer_tag is None:
            # Consume on the broadcast queue
            self._broadcast_consumer_tag = yield self._broadcast_queue.consume(self._on_broadcast)
        raise gen.Return(identifier)

    @gen.coroutine
    def remove_broadcast_subscriber(self, identifier):
        try:
            del self._broadcast_subscribers[identifier]
        except KeyError:
            raise ValueError("Broadcast subscriber '{}' unknown".format(identifier))
        if not self._broadcast_subscribers:
            yield self._broadcast_queue.cancel(self._broadcast_consumer_tag)
            self._broadcast_consumer_tag = None

    def channel(self):
        return self._channel

    @gen.coroutine
    def connect(self):
        """Get a channel and set up all the exchanges/queues we need"""
        if self._channel:
            # Already connected
            return

        exchange_params = copy.copy(EXCHANGE_PROPERTIES)

        if self._testing_mode:
            exchange_params.setdefault('auto_delete', self._testing_mode)

        self._channel = yield self._connection.channel()
        self._exchange = yield self._channel.declare_exchange(name=self._exchange_name, **exchange_params)

        yield self._create_broadcast_queue()

    @gen.coroutine
    def _create_broadcast_queue(self):
        """
        Create and bind the broadcast queue

        One is used for all broadcasts on this exchange
        """
        # Create a new such that we can see this is the broadcast queue
        name = "broadcast-{}".format(shortuuid.uuid())
        self._broadcast_queue = yield self._channel.declare_queue(
            name=name, exclusive=True, arguments=self._broadcast_queue_arguments)
        yield self._broadcast_queue.bind(self._exchange, routing_key=defaults.BROADCAST_TOPIC)

    @gen.coroutine
    def disconnect(self):
        yield self._channel.close()
        self._exchange = None
        self._channel = None

    @gen.coroutine
    def _on_rpc(self, subscriber, message):
        """
        :param subscriber: the subscriber function or coroutine that will get the RPC message
        :param message: the RMQ message
        :type message: :class:`topika.message.IncomingMessage`
        """
        with message.process(ignore_processed=True):
            # Tell the sender that we've dealt with it
            message.ack()
            msg = self._decode(message.body)

            try:
                receiver = utils.ensure_coroutine(subscriber)
                result = yield receiver(self, msg)
            except Exception as exception:  # pylint: disable=broad-except
                # We had an exception in  calling the receiver
                yield self._send_response(message.reply_to, message.correlation_id, utils.exception_response(exception))
            else:
                if concurrent.is_future(result):
                    yield self._send_future_response(result, message.reply_to, message.correlation_id)
                else:
                    # All good, send the response out
                    yield self._send_response(message.reply_to, message.correlation_id, utils.result_response(result))

    @gen.coroutine
    def _on_broadcast(self, message):
        with message.process():
            msg = self._decode(message.body)
            for receiver in self._broadcast_subscribers.values():
                try:
                    receiver = utils.ensure_coroutine(receiver)
                    yield receiver(self, msg[messages.BroadcastMessage.BODY], msg[messages.BroadcastMessage.SENDER],
                                   msg[messages.BroadcastMessage.SUBJECT],
                                   msg[messages.BroadcastMessage.CORRELATION_ID])
                except Exception:  # pylint: disable=broad-except
                    _LOGGER.exception('Exception in broadcast receiver')

    @gen.coroutine
    def _send_future_response(self, future, reply_to, correlation_id):
        """
        The RPC call returned a future which means we need to keep send a pending response
        and send a further message when the future resolves.  If it resolves to another future
        we should send out a further pending response and so on.

        :param future: the future from the RPC call
        :type future: :class:`tornado.concurrent.Future`
        :param reply_to: the recipient
        :param correlation_id: the correlation id
        """
        try:
            # Keep looping in case we're in a situation where a futrue resolves to a future etc.
            while concurrent.is_future(future):
                # Send out a message saying that we're waiting for a future to complete
                yield self._send_response(reply_to, correlation_id, utils.pending_response())
                future = yield future
        except kiwipy.CancelledError as exception:
            # Send out a cancelled response
            yield self._send_response(reply_to, correlation_id, utils.cancelled_response(str(exception)))
        except Exception as exception:  # pylint: disable=broad-except
            # Send out an exception response
            yield self._send_response(reply_to, correlation_id, utils.exception_response(exception))
        else:
            # We have a final result so send that as the response
            yield self._send_response(reply_to, correlation_id, utils.result_response(future))

    @gen.coroutine
    def _send_response(self, reply_to, correlation_id, response):
        assert reply_to, "Must provide an identifier for the recipient"

        message = topika.Message(body=self._response_encode(response), correlation_id=correlation_id)
        result = yield self._exchange.publish(message, routing_key=reply_to)
        raise gen.Return(result)


class RmqCommunicator(object):
    """
    A publisher and subscriber using topika and a tornado event loop
    """

    def __init__(self,
                 connection,
                 message_exchange=defaults.MESSAGE_EXCHANGE,
                 task_exchange=defaults.TASK_EXCHANGE,
                 task_queue=defaults.TASK_QUEUE,
                 task_prefetch_size=defaults.TASK_PREFETCH_SIZE,
                 task_prefetch_count=defaults.TASK_PREFETCH_COUNT,
                 queue_expires=defaults.QUEUE_EXPIRES,
                 encoder=defaults.ENCODER,
                 decoder=defaults.DECODER,
                 testing_mode=False):
        # pylint: disable=too-many-arguments
        """
        :param connection: The RMQ connector object
        :type connection: :class:`topika.Connection`
        :param message_exchange: The name of the RMQ message exchange to use
        :type message_exchange: str
        :param task_exchange: The name of the RMQ task exchange to use
        :type task_exchange: str
        :param task_queue: The name of the task queue to use
        :type task_queue: str
        :param queue_expires: the expiry time for standard queues in milliseconds.  This is the time after which, if
            there are no subscribers, a queue will automatically be deleted by RabbitMQ.
        :type queue_expires: int
        :param encoder: The encoder to call for encoding a message
        :param decoder: The decoder to call for decoding a message
        :param testing_mode: Run in testing mode: all queues and exchanges
            will be temporary
        """
        super(RmqCommunicator, self).__init__()

        self._connection = connection
        self._loop = connection.loop

        self._message_subscriber = RmqSubscriber(
            connection,
            message_exchange=message_exchange,
            queue_expires=queue_expires,
            encoder=encoder,
            decoder=decoder,
            testing_mode=testing_mode)
        self._message_publisher = RmqPublisher(
            connection, exchange_name=message_exchange, encoder=encoder, decoder=decoder, testing_mode=testing_mode)
        self._task_subscriber = tasks.RmqTaskSubscriber(
            connection,
            exchange_name=task_exchange,
            queue_name=task_queue,
            testing_mode=testing_mode,
            decoder=decoder,
            encoder=encoder,
            prefetch_size=task_prefetch_size,
            prefetch_count=task_prefetch_count)
        self._task_publisher = tasks.RmqTaskPublisher(
            connection,
            exchange_name=task_exchange,
            task_queue_name=task_queue,
            encoder=encoder,
            decoder=decoder,
            testing_mode=testing_mode,
        )

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def __str__(self):
        return "RMQCommunicator ({})".format(self._connection)

    @property
    def loop(self):
        return self._connection.loop

    @gen.coroutine
    def connect(self):
        if not self._connection.is_open:
            yield self._connection.connect()

        yield self._message_subscriber.connect()
        yield self._task_subscriber.connect()
        yield self._message_publisher.connect()
        yield self._task_publisher.connect()

    @gen.coroutine
    def disconnect(self):
        yield self._message_publisher.disconnect()
        yield self._message_subscriber.disconnect()
        yield self._task_publisher.disconnect()
        yield self._task_subscriber.disconnect()
        yield self._connection.close()

    @gen.coroutine
    def add_rpc_subscriber(self, subscriber, identifier=None):
        identifier = yield self._message_subscriber.add_rpc_subscriber(subscriber, identifier)
        raise gen.Return(identifier)

    @gen.coroutine
    def remove_rpc_subscriber(self, identifier):
        yield self._message_subscriber.remove_rpc_subscriber(identifier)

    def add_task_subscriber(self, subscriber):
        self._task_subscriber.add_task_subscriber(subscriber)

    def remove_task_subscriber(self, subscriber):
        self._task_subscriber.remove_task_subscriber(subscriber)

    @gen.coroutine
    def add_broadcast_subscriber(self, subscriber, identifier=None):
        identifier = yield self._message_subscriber.add_broadcast_subscriber(subscriber, identifier)
        raise gen.Return(identifier)

    @gen.coroutine
    def remove_broadcast_subscriber(self, identifier):
        yield self._message_subscriber.remove_broadcast_subscriber(identifier)

    @gen.coroutine
    def rpc_send(self, recipient_id, msg):
        """
        Initiate a remote procedure call on a recipient

        :param recipient_id: The recipient identifier
        :param msg: The body of the message
        :return: A future corresponding to the outcome of the call
        """
        try:
            response_future = yield self._message_publisher.rpc_send(recipient_id, msg)
            raise gen.Return(response_future)
        except pika.exceptions.UnroutableError as exception:
            raise kiwipy.UnroutableError(str(exception))

    @gen.coroutine
    def broadcast_send(self, body, sender=None, subject=None, correlation_id=None):
        result = yield self._message_publisher.broadcast_send(body, sender, subject, correlation_id)
        raise gen.Return(result)

    @gen.coroutine
    def task_send(self, task, no_reply=False):
        try:
            result = yield self._task_publisher.task_send(task, no_reply)
            raise gen.Return(result)
        except pika.exceptions.UnroutableError as exception:
            raise kiwipy.UnroutableError(str(exception))
        except pika.exceptions.NackError as exception:
            raise kiwipy.TaskRejected(str(exception))


class RmqThreadCommunicator(kiwipy.Communicator):
    """
    RabbitMQ communicator that runs an event loop on a separate thread to do communication.
    This also means that heartbeats are not missed and the main program is free to block for
    as long as it wants.
    """

    TASK_TIMEOUT = 5.

    @classmethod
    def connect(cls,
                connection_params=None,
                connection_factory=topika.connect_robust,
                loop=None,
                message_exchange=defaults.MESSAGE_EXCHANGE,
                task_exchange=defaults.TASK_EXCHANGE,
                task_queue=defaults.TASK_QUEUE,
                task_prefetch_size=defaults.TASK_PREFETCH_SIZE,
                task_prefetch_count=defaults.TASK_PREFETCH_COUNT,
                encoder=defaults.ENCODER,
                decoder=defaults.DECODER,
                testing_mode=False):
        # pylint: disable=too-many-arguments
        connection_params = connection_params or {}
        # Create a new loop if one isn't supplied
        loop = loop or ioloop.IOLoop()
        connection_params['loop'] = loop

        # Run the loop to create the connection
        connection = loop.run_sync(lambda: connection_factory(**connection_params))
        communicator = cls(
            connection,
            message_exchange=message_exchange,
            task_exchange=task_exchange,
            task_queue=task_queue,
            task_prefetch_size=task_prefetch_size,
            task_prefetch_count=task_prefetch_count,
            encoder=encoder,
            decoder=decoder,
            testing_mode=testing_mode)

        # Start the communicator
        communicator.start()
        return communicator

    def __init__(self,
                 connection,
                 message_exchange=defaults.MESSAGE_EXCHANGE,
                 task_exchange=defaults.TASK_EXCHANGE,
                 task_queue=defaults.TASK_QUEUE,
                 task_prefetch_size=defaults.TASK_PREFETCH_SIZE,
                 task_prefetch_count=defaults.TASK_PREFETCH_COUNT,
                 encoder=defaults.ENCODER,
                 decoder=defaults.DECODER,
                 testing_mode=False):
        # pylint: disable=too-many-arguments
        """
        :param connection: The RMQ connector object
        :type connection: :class:`topika.Connection`
        :param message_exchange: The name of the RMQ message exchange to use
        :type message_exchange: str
        :param task_exchange: The name of the RMQ task exchange to use
        :type task_exchange: str
        :param task_queue: The name of the task queue to use
        :type task_queue: str
        :param encoder: The encoder to call for encoding a message
        :param decoder: The decoder to call for decoding a message
        :param testing_mode: Run in testing mode: all queues and exchanges
            will be temporary
        """
        self._communicator = RmqCommunicator(
            connection,
            message_exchange=message_exchange,
            task_exchange=task_exchange,
            task_queue=task_queue,
            encoder=encoder,
            decoder=decoder,
            task_prefetch_size=task_prefetch_size,
            task_prefetch_count=task_prefetch_count,
            testing_mode=testing_mode)
        self._loop = self._communicator.loop
        self._communicator_thread = None

    def __enter__(self):
        self._ensure_running()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def loop(self):
        return self._loop

    def start(self):
        assert self._communicator_thread is None, "Already running"

        start_future = ThreadFuture()

        def run_loop():
            _LOGGER.debug('Starting event loop on %s', threading.current_thread())

            @gen.coroutine
            def do_connect():
                try:
                    yield self._communicator.connect()
                    start_future.set_result(True)
                except Exception as exception:  # pylint: disable=broad-except
                    import traceback
                    _LOGGER.error("Error starting the communicator:\n%s", traceback.format_exc())
                    start_future.set_exception(exception)

            self._loop.add_callback(do_connect)
            self._loop.start()
            self._communicator_thread = None

            _LOGGER.debug('Event loop stopped on %s', threading.current_thread())
            ioloop.IOLoop.current(False)

        self._communicator_thread = threading.Thread(
            target=run_loop, name="Communications thread for '{}'".format(self._communicator))
        self._communicator_thread.daemon = True
        self._communicator_thread.start()
        start_future.result()

    def stop(self):
        comm_thread = self._communicator_thread
        if comm_thread is None:
            return

        stop_future = ThreadFuture()

        @gen.coroutine
        def stop_loop():
            try:
                with kiwipy.capture_exceptions(stop_future):
                    yield self._communicator.disconnect()
            finally:
                self._loop.stop()
                stop_future.set_result(True)

        # The stop will end up setting self._communicator_thread to None
        self._loop.add_callback(stop_loop)
        comm_thread.join()
        stop_future.result()

    def add_rpc_subscriber(self, subscriber, identifier=None):
        coro = partial(self._communicator.add_rpc_subscriber, subscriber, identifier)
        return self._run_task(coro)

    def remove_rpc_subscriber(self, identifier):
        coro = partial(self._communicator.remove_rpc_subscriber, identifier)
        return self._run_task(coro)

    def add_task_subscriber(self, subscriber):
        self._communicator.add_task_subscriber(subscriber)

    def remove_task_subscriber(self, subscriber):
        self._communicator.remove_task_subscriber(subscriber)

    def add_broadcast_subscriber(self, subscriber, identifier=None):
        coro = partial(self._communicator.add_broadcast_subscriber, subscriber, identifier)
        return self._run_task(coro)

    def remove_broadcast_subscriber(self, identifier):
        coro = partial(self._communicator.remove_broadcast_subscriber, identifier)
        return self._run_task(coro)

    def task_send(self, task, no_reply=False):
        self._ensure_running()
        future = self._send_message(partial(self._communicator.task_send, task, no_reply))
        if no_reply:
            return None

        return self.tornado_to_kiwi_future(future)

    def rpc_send(self, recipient_id, msg):
        self._ensure_running()
        future = self._send_message(partial(self._communicator.rpc_send, recipient_id, msg))
        return self.tornado_to_kiwi_future(future)

    def broadcast_send(self, body, sender=None, subject=None, correlation_id=None):
        self._ensure_running()
        return self._send_message(
            partial(
                self._communicator.broadcast_send,
                body=body,
                sender=sender,
                subject=subject,
                correlation_id=correlation_id))

    def _send_message(self, coro):
        send_future = kiwipy.Future()

        @gen.coroutine
        def do_task():
            with kiwipy.capture_exceptions(send_future):
                send_future.set_result((yield coro()))

        self._loop.add_callback(do_task)
        return send_future.result()

    def tornado_to_kiwi_future(self, tornado_future):
        """
        Convert a tornado future to a kiwipy future

        :param tornado_future: the tornado future to convert
        :type tornado_future: :class:`tornado.concurrent.Future`
        :return: the kiwipy future
        :rtype: :class:`kiwipy.Future`
        """
        kiwi_future = kiwipy.Future()

        def done(done_future):
            # Copy over the future
            try:
                result = done_future.result()
                if concurrent.is_future(result):
                    # Change the future type to a kiwi one
                    result = self.tornado_to_kiwi_future(result)
                kiwi_future.set_result(result)
            except kiwipy.CancelledError:
                kiwi_future.cancel()
            except Exception as exception:  # pylint: disable=broad-except
                kiwi_future.set_exception(exception)

        self.loop().add_future(tornado_future, done)
        return kiwi_future

    def _create_task(self, coro):
        """
        Create a task to be run on our event loop

        :param coro: the coroutine to run
        :return: a future corresponding to the execution of the coroutine
        :rtype: :class:`tornado.concurrent.Future`
        """
        return utils.create_task(coro, self._loop)

    def _run_task(self, coro):
        """
        Run a coroutine on the event loop and return the result.  It may take a little time for the loop
        to get around to scheduling it so we use a timeout as set by the TASK_TIMEOUT class constant.

        :param coro: the coroutine to run
        :return: the result of running the coroutine
        """
        return self.tornado_to_kiwi_future(self._create_task(coro)).result(timeout=self.TASK_TIMEOUT)

    def _ensure_running(self):
        if self._communicator_thread is not None:
            return
        self.start()


def connect(connection_params=None,
            connection_factory=topika.connect_robust,
            loop=None,
            message_exchange=defaults.MESSAGE_EXCHANGE,
            task_exchange=defaults.TASK_EXCHANGE,
            task_queue=defaults.TASK_QUEUE,
            task_prefetch_size=defaults.TASK_PREFETCH_SIZE,
            task_prefetch_count=defaults.TASK_PREFETCH_COUNT,
            encoder=defaults.ENCODER,
            decoder=defaults.DECODER,
            testing_mode=False):
    # pylint: disable=too-many-arguments
    return RmqThreadCommunicator.connect(
        connection_params=connection_params,
        connection_factory=connection_factory,
        loop=loop,
        message_exchange=message_exchange,
        task_exchange=task_exchange,
        task_queue=task_queue,
        task_prefetch_size=task_prefetch_size,
        task_prefetch_count=task_prefetch_count,
        encoder=encoder,
        decoder=decoder,
        testing_mode=testing_mode)



from __future__ import absolute_import
import topika
from tornado import testing, gen
import unittest
import uuid

import kiwipy.rmq

try:
    import pika
    import pika.exceptions
    from kiwipy import rmq
except ImportError:
    pika = None


@unittest.skipIf(not pika, "Requires pika library and RabbitMQ")
class TestTaskPublisher(testing.AsyncTestCase):

    def setUp(self):
        super(TestTaskPublisher, self).setUp()
        self.loop = self.io_loop

        exchange_name = "{}.{}".format(self.__class__.__name__, uuid.uuid4())
        task_queue_name = "{}.{}".format(self.__class__.__name__, uuid.uuid4())

        self.connection = None  # type: topika.Connection
        self.task_publisher = None  # type: kiwipy.rmq.RmqTaskPublisher

        @gen.coroutine
        def init():
            self.connection = yield topika.connect_robust('amqp://guest:guest@localhost:5672/', loop=self.loop)
            self.task_publisher = rmq.RmqTaskPublisher(
                self.connection, task_queue_name=task_queue_name, exchange_name=exchange_name, testing_mode=True)
            yield self.task_publisher.connect()

        self.loop.run_sync(init)

    def tearDown(self):

        @gen.coroutine
        def destroy():
            yield self.task_publisher.disconnect()
            yield self.connection.close()

        self.loop.run_sync(destroy)

        # Call this last, it closes the loop
        super(TestTaskPublisher, self).tearDown()

    @testing.gen_test
    def test_send_no_subscribers(self):
        """ Test what happens when there are no task queues bound to the exchange """
        TASK = 'The meaning?'
        with self.assertRaises(pika.exceptions.UnroutableError):
            result = yield self.task_publisher.task_send(TASK)
